"""Displays summary of what will be run using Rich."""
from rich import print
from rich.columns import Columns
from rich.prompt import Confirm

from .all_panes import AllPanes, PipelineSelector


def show_summary(selector: PipelineSelector):
    """
    Display summary of what will be run using Rich.

    Will ask confirmation of whether to continue.
    """
    if len(selector.all_jobs[0]) == 0:
        if len(selector.selected_files) == 0:
            print("No output files were selected, so the pipeline cannot be run.\n")
        else:
            print("All output files for selected templates already exists.\n")
        if Confirm.ask(
            "Do you want to return to the selector pane to select output files?"
        ):
            return AllPanes.SELECTOR
        return None

    print("[b]Selected output files[/b]")
    print(Columns(sorted(selector.selected_files)))

    print("")
    print("[b]Pipeline report[/b]")
    selector.job_list.report()

    print("")
    print("[b]Jobs[/b]")
    print(Columns(sorted([str(job) for job in selector.all_jobs[0]])))
    if Confirm.ask("Do you want to run this pipeline?"):
        return AllPanes.RUN
    return AllPanes.SELECTOR



"""Defines enumeration of all available panes for user to select the next one."""
from enum import Enum

from file_tree import FileTree
from fsl_pipe import Pipeline
from fsl_pipe.job import FileTarget, JobList, RunMethod, get_target


class AllPanes(Enum):
    """
    Enumeration of all the possible GUI panes.

    Panes should return one of these to indicate what the next pane will be.
    If None is returned by a pane instead, the application will quit.
    """

    PLACEHOLDER = 1
    SELECTOR = 2
    SUMMARY = 3
    RUN = 4


class PipelineSelector:
    """
    Define which parts of the pipeline will be run.

    This information will be updated by each pane in the pipeline.
    """

    def __init__(
        self,
        pipeline: Pipeline,
        file_tree: FileTree,
        overwrite_dependencies=False,
        run_method=None,
    ):
        """Create new pipeline selector to be edited in GUI."""
        self.file_tree = file_tree
        all_templates = file_tree.template_keys()
        templates = {
            elem
            for piped_function in pipeline.scripts
            for elem in set.union(
                piped_function.filter_templates(False, all_templates),
                piped_function.filter_templates(True, all_templates),
            )
        }
        self.partial_tree = file_tree.filter_templates(templates)
        self.pipeline = pipeline
        self.full_job_list = self.pipeline.generate_jobs(self.file_tree)
        self.selected_files = set()
        self.all_jobs = ([], set())

        self._overwrite_dependencies = overwrite_dependencies
        if run_method is None:
            run_method = RunMethod.default()
        self.run_method = run_method

    @property
    def overwrite_dependencies(
        self,
    ):
        """Whether to override any already existing dependencies."""
        return self._overwrite_dependencies

    @overwrite_dependencies.setter
    def overwrite_dependencies(self, value):
        """Ensure jobs are reset when overwrite_dependencies is updated."""
        self._overwrite_dependencies = value
        self.reset_all_jobs()

    def get_file_targets(self, path) -> FileTarget:
        """Return file target corresponding to path."""
        return get_target(path, self.full_job_list.targets)

    def update_placeholders(self, **new_placeholders):
        """Update full job list for if placeholder values got changed."""
        self.file_tree.placeholders.update(**new_placeholders)
        self.partial_tree.placeholders.update(**new_placeholders)
        self.full_job_list = self.pipeline.generate_jobs(self.file_tree)
        self.reset_all_jobs()

    def add_selected_file(self, file):
        """Add a specific file to the list to be produced."""
        self.selected_files.add(file)
        target = self.get_file_targets(file)
        target.producer.add_to_jobs(
            self.all_jobs,
            overwrite=True,
            overwrite_dependencies=self.overwrite_dependencies,
        )

    def remove_selected_file(self, file):
        """Add a specific file to the list to be produced."""
        self.selected_files.remove(file)
        self.reset_all_jobs()

    def reset_all_jobs(
        self,
    ):
        """Force a recalculation of all_jobs."""
        self.all_jobs = ([], set())
        for file in self.selected_files:
            if (
                self.get_file_targets(file).producer is not None
                and self.get_file_targets(file).producer.expected()
            ):
                self.add_selected_file(file)

    @property
    def job_list(
        self,
    ) -> JobList:
        """Collection of jobs that will be run."""
        return JobList(self.file_tree, self.all_jobs[0], self.full_job_list.targets)

    def run(self):
        """Run the pipeline."""
        return self.job_list.run(self.run_method)



"""
Textual tab allowing the editing of placeholder values.

The main goal of this tab is to allow the placeholder values
to be changed from that included within the user-provied `FileTree`.
These changes are immediately made into the `FileTree`.
"""
from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.widgets import Button, Footer, Header, Input, Label

from .all_panes import AllPanes, PipelineSelector


class PlaceholderEditor(App):
    """
    GUI to view pipeline.

    There are two separate pages (visiblity set by `show_page`):
    - `placeholder_tab` allows the user to change the placeholder values.
    - `output_tab` allows the user to select the requested output files.
    """

    TITLE = "FSL pipeline"

    def __init__(self, selector: PipelineSelector):
        """Create the pipeline GUI."""
        super().__init__()
        self.selector = selector
        file_tree = selector.file_tree
        self.keys = [Label(k) for k in file_tree.placeholders.keys()]
        self.values = [
            Input(", ".join(v) if isinstance(v, list) else v)
            for v in file_tree.placeholders.values()
        ]

    def compose(self) -> ComposeResult:
        """Build the pipeline GUI."""
        yield Header()
        yield Vertical(
            Label("step 1: optionally edit placeholder values"),
            Horizontal(
                Vertical(Label("Placeholders"), *self.keys),
                Vertical(Label("Values"), *self.values),
            ),
            Button("Continue"),
        )
        yield Footer()

    def on_button_pressed(self, event: Button.Pressed):
        """Continue to the next app."""
        new_placeholders = {}
        for key, value in zip(self.keys, self.values):
            text = value.value
            if "," in text:
                new_value = [elem.strip() for elem in text.split(",")]
            else:
                new_value = text.strip()
            new_placeholders[str(key.renderable)] = new_value

        self.selector.update_placeholders(**new_placeholders)
        self.exit(AllPanes.SELECTOR)



"""Visualise the output selection part of the FSL pipeline GUI."""
import itertools
import os.path as op
from functools import lru_cache

from file_tree import FileTree, Template
from fsl_pipe.job import RunMethod
from rich.style import Style
from rich.text import Text
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.message import Message
from textual.widgets import Button, Checkbox, DataTable, Footer, Header, Select, Tree
from textual.widgets.tree import TreeNode

from .all_panes import AllPanes, PipelineSelector


class TemplateSelect(Message, bubble=True):
    """Message sent when a template in the sidebar gets selected."""

    def __init__(self, sender, template: Template):
        """Create template selector."""
        self.template = template
        super().__init__(sender)


class OutputSelector(App):
    """Textual application to select pipeline output templates/files."""

    TITLE = "FSL pipeline"
    CSS_PATH = "css/selector.css"

    BINDINGS = [
        Binding("r", "run", "Run pipeline", show=True),
        Binding("a", "select_all", "Select all for this template", show=True),
    ]

    def __init__(self, selector: PipelineSelector):
        """Create pipeline output selector."""
        super().__init__()
        self.selector = selector

    @property
    def file_tree(
        self,
    ):
        """Input/output file structure."""
        return self.selector.file_tree

    @property
    def pipeline(
        self,
    ):
        """Set of pipeline recipes."""
        return self.selector.pipeline

    def compose(self) -> ComposeResult:
        """Create basic layout."""
        yield Header()
        self.template_renderer = TemplateRenderer(self.selector)
        yield Vertical(
            Horizontal(
                TemplateTreeControl(self.selector.partial_tree, self.template_renderer),
                self.template_renderer,
            ),
            Horizontal(
                Button("Back", name="back"),
                Button("Run pipeline", name="run"),
                Select(
                    [(e.name, e) for e in RunMethod], value=self.selector.run_method
                ),
                Checkbox(
                    "Overwrite dependencies", value=self.selector.overwrite_dependencies
                ),
                id="control-bar",
            ),
        )
        yield Footer()

    async def handle_template_select(self, message: TemplateSelect):
        """User has selected a template."""
        template = message.template
        self.app.sub_title = template.as_string
        await self.body.update(TemplateRenderer(template, self.file_tree))

    def on_button_pressed(self, event: Button.Pressed):
        """Continue to the next app."""
        if event.button.name == "back":
            self.exit(AllPanes.PLACEHOLDER)
        elif event.button.name == "run":
            self.action_run()

    def on_checkbox_changed(self, event: Checkbox.Changed):
        """Update overwrite dependencies."""
        self.selector.overwrite_dependencies = event.checkbox.value

    def on_select_changed(self, event: Select.Changed):
        """Update run method."""
        self.selector.run_method = event.select.value

    def action_run(
        self,
    ):
        """Run the pipeline."""
        self.exit(AllPanes.SUMMARY)

    def action_select_all(self):
        """Select all files matching this template."""
        self.template_renderer.select_all()


class TemplateTreeControl(Tree):
    """Sidebar containing all template definitions in FileTree."""

    current_node = None
    BINDINGS = [
        Binding("space", "toggle_node", "Collapse/Expand Node", show=True),
        Binding("up", "cursor_up", "Move Up", show=True),
        Binding("down", "cursor_down", "Move Down", show=True),
    ]

    def __init__(self, file_tree: FileTree, renderer, name: str = None):
        """
        Create a new template sidebar based on given FileTree.

        :param file_tree: FileTree to interact with
        :param renderer: Panel showing the files corresponding to selected template.
        :param name: name of the sidebar within textual.
        """
        self.file_tree = file_tree
        super().__init__("-", name=name)
        self.show_root = False
        self.find_children(self.root, self.file_tree.get_template(""))
        self.root.expand_all()
        self.renderer = renderer
        self.select_node(self.get_node_at_line(0))

    def on_mount(
        self,
    ):
        """Take focus on mount."""
        self.focus()

    def find_children(self, parent_node: TreeNode, template: Template):
        """
        Find all the children of a template and add them to the node.

        Calls itself recursively.
        """
        all_children = template.children(self.file_tree._templates.values())
        if len(all_children) == 0:
            parent_node.add_leaf(template.unique_part, template)
        else:
            this_node = parent_node.add(template.unique_part, template)
        children = set()
        for child in all_children:
            if child not in children:
                self.find_children(this_node, child)
                children.add(child)

    def render_label(self, node: TreeNode[Template], base_style, style):
        """Render line in tree."""
        if node.data is None:
            return node.label
        label = _render_node_helper(self.file_tree, node).copy()
        if node is self.cursor_node:
            label.stylize("reverse")
        if not node.is_expanded and len(node.children) > 0:
            label = Text("📁 ") + label
        return label

    def on_tree_node_highlighted(self):
        """Inform other panel if template is selected."""
        if self.current_node is not self.cursor_node:
            self.current_node = self.cursor_node
            self.renderer.render_template(self.current_node.data)


@lru_cache(None)
def _render_node_helper(tree: FileTree, node: TreeNode[Template]):
    meta = {
        "@click": f"click_label({node.id})",
        "tree_node": node.id,
        # "cursor": node.is_cursor,
    }
    paths = tree.get_mult(
        _get_template_key(tree, node.data), filter=True
    ).data.flatten()
    existing = [p for p in paths if p != ""]
    color = "blue" if len(existing) == len(paths) else "yellow"
    if len(existing) == 0:
        color = "red"
    counter = f" [{color}][{len(existing)}/{len(paths)}][/{color}]"
    res = Text.from_markup(
        node.data.rich_line(tree._templates) + counter, overflow="ellipsis"
    )
    res.apply_meta(meta)
    return res


def _get_template_key(tree, template):
    """Get key representing template with file-tree."""
    keys = {k for k, t in tree._templates.items() if t is template}
    return next(iter(keys))


class TemplateRenderer(DataTable):
    """
    Helper class to create a Rich rendering of a template.

    There are two parts:

        - a text file with the template
        - a table with the possible placeholder value combinations
          (shaded red for non-existing files)
    """

    def __init__(self, selector: PipelineSelector):
        """Create new renderer for template."""
        super().__init__()
        self.selector = selector
        self.cursor_type = "row"

    @property
    def file_tree(
        self,
    ):
        """Input/output file structure."""
        return self.selector.file_tree

    def on_mount(self):
        """Render upper-level template on mount."""
        self.render_template(self.file_tree.get_template(""))

    def render_template(self, template: Template):
        """Render the template as rich text."""
        self.current_template = template
        self.clear(columns=True)
        xr = self.file_tree.get_mult(
            _get_template_key(self.file_tree, template), filter=False
        )
        coords = sorted(xr.coords.keys())
        self.add_column("", key="checkboxes")
        self.add_columns("", *coords, "filename")
        for values in itertools.product(*[xr.coords[c].data for c in coords]):
            path = xr.sel(**{c: v for c, v in zip(coords, values)}).item()
            style = Style(bgcolor="blue" if op.exists(path) else None)
            self.add_row(
                self.get_checkbox(path),
                *[Text(v, style=style) for v in values],
                Text(path, style=style),
                key=path,
            )

    def on_data_table_row_selected(self, message: DataTable.RowSelected):
        """Add or remove selected row from pipeline run."""
        self.toggle_checkbox(message.row_key)

    def select_all(
        self,
    ):
        """Add or remove selected row from pipeline run."""
        for row in self.rows:
            self.toggle_checkbox(row, True)

    def toggle_checkbox(self, row_key, new_value=None):
        """Toggles the checkbox in given row."""
        current_checkbox = self.get_cell(row_key, "checkboxes")
        if current_checkbox == "☑" and new_value is not True:
            self.selector.remove_selected_file(row_key.value)
            self.update_cell(row_key, "checkboxes", self.get_checkbox(row_key.value))
        elif current_checkbox in "☐⊡" and new_value is not False:
            self.selector.add_selected_file(row_key.value)
            self.update_cell(row_key, "checkboxes", "☑")

    def get_checkbox(self, filename):
        """Return checkbox corresponding to filename."""
        if filename in self.selector.selected_files:
            return "☑"
        target = self.selector.get_file_targets(filename)
        if target.producer is None:
            return ""
        elif target.producer in self.selector.all_jobs[1]:
            return "⊡"
        elif target.producer.expected():
            return "☐"
        else:
            return "M"



"""Main module containing the actual code to run the GUI."""
from file_tree import FileTree
from fsl_pipe import Pipeline
from rich import print

from .all_panes import AllPanes, PipelineSelector
from .placeholder_tab import PlaceholderEditor
from .selector_tab import OutputSelector
from .summary import show_summary


def run_gui(pipeline: Pipeline, tree: FileTree, **kwargs):
    """
    Run terminal-based GUI, so the user can select which output files to produce.

    Returns True if a pipeline actually ran. False if the user exited before running a pipeline.

    :param pipeline: Collection of recipes contained in an `fsl-pipe` Pipeline object.
    :param tree: Description of the directory structure containing the input and output files.
    """
    selector = PipelineSelector(pipeline, tree, **kwargs)
    current = AllPanes.PLACEHOLDER if len(tree.placeholders) > 0 else AllPanes.SELECTOR

    while True:
        if current is None:
            print("Exiting FSL pipeline output selector")
            return False
        elif current == AllPanes.PLACEHOLDER:
            current = PlaceholderEditor(selector).run()
        elif current == AllPanes.SELECTOR:
            current = OutputSelector(selector).run()
        elif current == AllPanes.SUMMARY:
            current = show_summary(selector)
        elif current == AllPanes.RUN:
            selector.run()
            return True
        else:
            raise ValueError(
                f"Unexpected error: next pane is set to unknown value: {current}"
            )
